/*
 * Copyright (c) 2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <libusb.h>

#include <daqflex/daqflex.h>

#include "daqflexInt.h"

static int isMccDevice(struct libusb_device_descriptor *);
static int getScanParams(struct mccdevice *);
static unsigned short getBulkPacketSize(uint8_t *, int);
static uint8_t getEndPointInAddress(uint8_t *, int);
static uint8_t getEndPointOutAddress(uint8_t *, int);

#define MIN(a,b) ((a)<(b)?(a):(b))

/* for older libusb */
#ifndef HAVE_LIBUSB_ERROR_NAME
static char *
libusb_error_name(int err)
{
	static char buf[100];
	snprintf(buf, sizeof(buf), "error %d\n", abs(err));
	return buf;
}
#endif

/**
 * Is the USB device a supported MCC One
 *
 * XXX for now only claim to support USB-201 devices
 */
static int
isMccDevice(struct libusb_device_descriptor *desc)
{
	if (desc->idVendor != MCC_VENDOR_ID)
		return 0;
	switch (desc->idProduct) {
	case MCC_PRODUCT_USB_201:
		return 1;
	default:
		return 0;
	}
}

/** 
 * Get USB Endpoints and Bulk transfer size for the device 
 */
static int
getScanParams(struct mccdevice *dev)
{
	int len;
	uint8_t endPoint[MAX_MESSAGE_LENGTH];

	len = libusb_control_transfer(dev->usbDev, STRINGMESSAGE,
	    LIBUSB_REQUEST_GET_DESCRIPTOR, LIBUSB_DT_CONFIG << 8, 0, endPoint,
	    sizeof(endPoint), DAQFLEX_USB_TIMEOUT);
	if (len < 0) {
		fprintf(stderr, "getScanParams: libusb_control_transfer() "
		    "failed: %s\n", libusb_error_name(len));
		return -1;
	}
	dev->inEndPoint = getEndPointInAddress(endPoint, len);
	dev->outEndPoint = getEndPointOutAddress(endPoint, len);
	dev->bulkPacketSize = getBulkPacketSize(endPoint, len);
	return 0;
}

/**
 * Parse USB descriptor and return Bulk transfer size
 */
static unsigned short
getBulkPacketSize(uint8_t *data, int len)
{
	int descriptorType;
	int length;
	int index = 0;

	while (1) {
		length = data[index];
		descriptorType = data[index+1];
		if (length == 0)
			break;
		if (descriptorType != 0x05) {
			index += length;
		} else {
			if ((data[index+2] & 0x80) != 0) {
				/* found packet size */
				return (unsigned short)
				    (data[index+5] << 8 | data[index + 4]);
			} else
				index += length;
		}
		if (index >= len)
			break;
	}
	return 0;
}

/**
 * Parse USB descriptor and return input end point
 */
static uint8_t
getEndPointInAddress(uint8_t *data, int len)
{
	int descriptorType;
	int length;
	int index = 0;

	while (1) {
		length = data[index];
		descriptorType = data[index + 1];

		if (length == 0)
			break;

		if (descriptorType != LIBUSB_DT_ENDPOINT) {
			index += length;
		} else {
			if ((data[index + 2] & LIBUSB_ENDPOINT_DIR_MASK)
			    == LIBUSB_ENDPOINT_IN)
				return data[index + 2];
			else
				index += length;
		}
		if (index >= len)
			break;
	}
	return 0;
}

/**
 * Parse USB descriptor and return output end point
 */
static uint8_t
getEndPointOutAddress(uint8_t* data, int data_length)
{
	int descriptorType;
	int length;
	int index = 0;

	while (1) {
		length = data[index];
		descriptorType = data[index + 1];

		if (length == 0)
			break;

		if (descriptorType != LIBUSB_DT_ENDPOINT) {
			index += length;
		} else {
			if ((data[index + 2] & LIBUSB_ENDPOINT_DIR_MASK)
			== LIBUSB_ENDPOINT_OUT)
				return data[index + 2];
			else
				index += length;
		}
		if (index >= data_length)
			break;
	}
	return 0;
}

/** 
 * daqflexOpen - open a connextion to a MCC USB device
 *
 * XXX currently assume that there is only one such device connected. 
 */
struct mccdevice *
daqflexOpen(void)
{
	int i, retval;
	ssize_t len;
	libusb_device **devices, *dev;
	struct libusb_device_descriptor desc;
	libusb_device_handle *handle;
	struct mccdevice *mcc;

	mcc = (struct mccdevice *)calloc(1, sizeof(struct mccdevice));
	if (mcc == NULL) {
		fprintf(stderr, "daqflexOpen: malloc() failed\n");
		return NULL;
	}

	retval = libusb_init(NULL);
	if (retval != 0) {
		fprintf(stderr, "daqflexOpen: libusb_init() failed: %s\n",
		    libusb_error_name(retval));
		free(mcc);
		return NULL;
	}
	len = libusb_get_device_list(NULL, &devices);
	if (len < 0) {
		fprintf(stderr, "daqflexOpen: libusb_get_device_list()"
		    " failed: %s\n", libusb_error_name(len));
		libusb_exit(NULL);
		free(mcc);
		return NULL;
	}
	if (len == 0)
		goto notfound;
	for (i = 0; i < len; i++) {
		dev = devices[i];
		libusb_get_device_descriptor(dev, &desc);
		if (isMccDevice(&desc))
			break;
	}
	if (i == len)
		goto notfound;
	/* found device */
	retval = libusb_open(dev, &handle);
	if (retval != LIBUSB_SUCCESS) {
		fprintf(stderr, "daqflexOpen: libusb_open() failed: %s\n",
		    libusb_error_name(retval));
		libusb_exit(NULL);
		free(mcc);
		return NULL;
	}
	retval = libusb_claim_interface(handle, 0);
	if (retval != LIBUSB_SUCCESS) {
		fprintf(stderr, "daqflexOpen: libusb_claim_interface()"
		    " failed: %s\n", libusb_error_name(retval));
		libusb_close(handle);
		libusb_exit(NULL);
		free(mcc);
		return NULL;
	}
	mcc->usbDev = handle;
	mcc->devices = devices;
	if (getScanParams(mcc) != 0) {
		daqflexClose(mcc);
		return NULL;
	}
	return mcc;

notfound:
	fprintf(stderr, "daqflexOpen: can't find device\n");
	libusb_exit(NULL);
	free(mcc);
	return NULL;
}

/**
 * daqflexClose - terminate the connection to the device 
 *
 * and free resources
 */
void
daqflexClose(struct mccdevice *dev)
{
	libusb_release_interface(dev->usbDev, 0);
	libusb_close(dev->usbDev);
	libusb_free_device_list(dev->devices, 1);
	libusb_exit(NULL);
	free(dev);
}

/**
 * do a control tranfer to the device
 *
 * should call getControlTransfer() shortly after calling that to read
 * the message sent back by the device
 */
static int
sendControlTransfer(struct mccdevice *dev, char *message)
{
	int i, len;
	size_t msglen = strlen(message);
	uint8_t data[MAX_MESSAGE_LENGTH];

	if (msglen > MAX_MESSAGE_LENGTH)
		return -1;

	memset(data, 0, MAX_MESSAGE_LENGTH);
	for (i = 0; i < msglen; i++)
		data[i] = toupper(message[i]);
	len = libusb_control_transfer(dev->usbDev,
	    LIBUSB_REQUEST_TYPE_VENDOR + LIBUSB_ENDPOINT_OUT,
	    STRINGMESSAGE, 0, 0, data, MAX_MESSAGE_LENGTH,
	    DAQFLEX_USB_TIMEOUT);
	if (len < 0) {
		fprintf(stderr, "libusb_control_transfer(out) failed: %s\n",
		    libusb_error_name(len));
		return -1;
	}
	return len;
}

/**
 * do a control transfer from the device
 * need to call sendControlTransfer() first to request data
 */
static int
getControlTransfer(struct mccdevice *dev, char *message, int len)
{
	int msglen;
	uint8_t data[64];

	msglen = libusb_control_transfer(dev->usbDev,
	    LIBUSB_REQUEST_TYPE_VENDOR + LIBUSB_ENDPOINT_IN,
	    STRINGMESSAGE, 0, 0, data, MAX_MESSAGE_LENGTH,
	    DAQFLEX_USB_TIMEOUT);
	if (msglen < 0) {
		fprintf(stderr, "libusb_control_transfer(in) failed: %s\n",
		    libusb_error_name(msglen));
		return -1;
	}
	memcpy(message, data, MIN(len, msglen));
	return MIN(len, msglen);
}

/**
 * daqflexSendMessage() - full control transfer to/from the device
 *
 * dev - device descriptor
 * message - message sent to the device
 * response - buffer to hold the answer from the device
 * maxlen - allocated length of 'response'
 *
 * returns the number of bytes actually read from the device.
 */
int
daqflexSendMessage(struct mccdevice *dev, char *message, char *response, int maxlen)
{
	int retval;

	retval = sendControlTransfer(dev, message);
	if (retval < 0)
		return -1;
	retval = getControlTransfer(dev, response, maxlen);
	if (retval < 0)
		return -1;
	return retval;
}

/**
 * daqflexFlushInputData - flush input data
 *
 * dev - device descriptor 
 *
 * Empties the USB buffers from pending data
 */
void
daqflexFlushInputData(struct mccdevice *dev)
{
	int len = 0;
	int status;
	uint8_t buf[dev->bulkPacketSize];

	do {
		status = libusb_bulk_transfer(dev->usbDev, dev->inEndPoint,
		    buf, sizeof(buf), &len, DAQFLEX_USB_TIMEOUT);
	} while (len > 0 && status == LIBUSB_SUCCESS);
}

/**
 * read an USB packet in bulk mode from the device
 */
int
daqflexReadPacket(struct mccdevice *dev, unsigned char *data, int len)
{
	int retval, transferred;

	if (len < dev->bulkPacketSize)
		return -1;
	retval = libusb_bulk_transfer(dev->usbDev,
	    dev->inEndPoint, (uint8_t *)data, dev->bulkPacketSize,
	    &transferred, DAQFLEX_USB_TIMEOUT);
	// printf("retval %d transferred %d\n", retval, transferred);
	if (retval < 0) {
		fprintf(stderr, "bulk transfer(in) failed: %s\n",
		    libusb_error_name(retval));
		return -1;
	}
	return transferred;
}

